function [multi,ind,J] = quaderror_clustering(n,f_of_X,K,sigma,info_sigma,seed,method)

% Creates a multi-task problem, then solves it with the assumption that the
% functions lie in two clusters. Computes the quadratic error of the
% multi-task estimator. Also computes the quadratic error of  the
% independent (single-task) procedure.  
%
% [multi,ind,J] = quaderror_clustering(n,d,m,I,sigma,info_sigma,seed,method,link)
%
% Input
% n : number of observations
% f_of_X : matrix of f^j(X_i). Row : observations, columns : tasks. size n*p
% K : kernel matrix. size n*n
% sigma : covariance matrix. size p*p
% info_sigma : string with the known information on sigma.
%   - 'nothing' : entirely estimates sigma.
%   - 'diagonal' : supposes that sigma is diagonal.
%   - 'isotropic' : supposes that sigma is of the form s^2*I_n.
%   - 'known' : supposes sigma known.
% seed : integer, used to initiate the pseudo-random generator.
% method : string which indicates how to estimate the variance.
%   - 'half' : takes the smallest index such that df < L/2 (recommended)
%   - 'largest_jump' : takes the largest jump of df.
%
% Output
% multi : quadratic error of the multi-task estimator
% ind : quadratic error of the single-task estimator
% J : selected clustering. size p*1
%
% Matrix used, for a cluster I of size k :
% M = lambda*I_p + mu*Diag(I) +
% mu*Diag(Ic)-mu*1_I*1_I'/k-mu*1_Ic*1_Ic'/(p-k) .

warning off;
rand('state',seed);
randn('state',seed);
warning on;

p = size(sigma,1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Data generation

fX = column(f_of_X);
V_sigma = kron(sigma,eye(n));
c_sigma = chol(V_sigma)';
y = fX + c_sigma*randn(n*p,1);
compY = compact(y,n,p);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Covariance matrix estimation

[uK,DK] = eig(K);
eK = max( real( diag(DK) ),0 );
[~,bK] = sort(eK,'descend');
eK = eK(bK);
uK = uK(:,bK);


if  strcmp(info_sigma,'nothing'),      
    % g : vector of pen_min(l)
    g = (0:n)';
    % df : vector of trace(A_l)
    dfs= 0:n;
    lambdas = dfs;
    % f : matrix of ||Y_z-A_l*Y_z||/n, with z in Z, the different z being
    % ordered in the columns as in the article : first the e_i, then the e_i+e_j.
    f = zeros(length(dfs),p*(p+1)/2);
    % Computation of all A_l, df(l) and g(l)
    for i = 1:length(dfs)
        if dfs(i)>0
            lambdas(i) = df_to_lambda_eig(dfs(i),eK);
        else
            lambdas(i) = Inf;
        end
        A = uK* diag( eK./(eK + n*p*lambdas(i)*ones(n,1)) )*uK';
        g(i) = (2*trace(A)-trace(A'*A))/(n*p);
        for k = 1:p
            z = ((1:p)' == k);
            Y = compY*z;
            ecart = (eye(n)-A)*Y;
            f(i,k) = trace( ecart'* ecart)/(n*p);
        end    
        for k = 1:(p-1)
            for j = (k+1):p
                z = ((1:p)' == k) + ((1:p)' == j);
                Y = compY*z;
                ecart = (eye(n)-A)*Y;
                f(i,p+(k-1)*p+j-k*(k+1)/2) =   trace( ecart'*ecart)/(n*p);
            end    
        end   
    end
elseif strcmp(info_sigma,'diagonal'),   
       g = (0:n)'; 
       dfs= 0:n;
       lambdas = dfs;              
       f = zeros(length(dfs),p);
       for i = 1:length(dfs)
           if dfs(i)>0
              lambdas(i) = df_to_lambda_eig(dfs(i),eK);
           else
              lambdas(i) = Inf;
           end
           A = uK* diag( eK./(eK + n*p*lambdas(i)*ones(n,1)) )*uK';
           g(i) = (2*trace(A)-trace(A'*A))/(n*p);
           for k = 1:p
               z = ((1:p)' == k);
               Y = compY*z;
               ecart = (eye(n)-A)*Y;
               f(i,k) = trace( ecart'* ecart)/(n*p);
           end     
       end
elseif strcmp(info_sigma,'isotropic'),
       [uK_i,eK_i] = eig_kron(eye(p),ones(p,1),uK,eK);
       [~,bK_i] = sort(eK_i,'descend');
       eK_i = eK_i(bK_i);
       uK_i = uK_i(:,bK_i);
       g = (0:(n*p))'; 
       dfs= (0:(n*p));
       lambdas = dfs;              
       f = zeros(length(dfs),1);
       for i = 1:length(dfs)
           if dfs(i)>0
              lambdas(i) = df_to_lambda_eig(dfs(i),eK_i);
           else
              lambdas(i) = Inf;
           end
           A = uK_i* diag( eK_i./(eK_i + n*p*lambdas(i)*ones(n*p,1)) )*uK_i'; 
           g(i) = (2*trace(A)-trace(A'*A))/(n*p);
           ecart = (eye(n*p)-A)*y;
           f(i,1) = trace( ecart'* ecart)/(n*p);
       end
end      
    
% Estimation of sigma : S
if strcmp(info_sigma,'nothing'),
    S = zeros(p,p);
    for k = 1:p
        S(k,k) = single_task(f(:,k),dfs,g,method);
    end
    for k = 1:(p-1)
        for j = (k+1):p
            x = single_task(f(:,p+(k-1)*p+j-k*(k+1)/2),dfs,g,method);
            S(k,j) = (x-S(k,k)-S(j,j))/2;
            S(j,k) = S(k,j);
        end
    end
elseif strcmp(info_sigma,'diagonal'),
       S = zeros(p,p);
       for k = 1:p
           S(k,k) = single_task(f(:,k),dfs,g,method);
       end
elseif strcmp(info_sigma,'isotropic'),
       s =  single_task(f,dfs,g,method);
       S = s*eye(p);
elseif strcmp(info_sigma,'known') 
       S = sigma;
else S = NaN;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Selection of the partition / model J

crit_J = zeros(1,2^p-1);
mus_J = zeros(2,2^p-1);

% Eigen-elements of the kernels of dimensions 2 and p-2

[~,eK_1] = eig_kron(eye(2),ones(2,1),uK',eK);
[~,bK_1] = sort(eK_1,'descend');
eK_1 = eK_1(bK_1);

[~,eK_2] = eig_kron(eye(p-2),ones(p-2,1),uK',eK);
[~,bK_2] = sort(eK_2,'descend');
eK_2 = eK_2(bK_2);

% Exploration of the partitions of {1,...,p}

for j = 1:(2^p-2) %2^p-1 : J = {1,...,p}
    J = str2num(sscanf(dec2bin(j,p),'%c',[p,1]));
    Jc = ones(p,1)-J;% insures that J is a scalar, not a boolean
    k = sum(J);
    
% Since the matrices M(mu_1,mu_2) can be jointly diagonalized, it is
% equivalent to two concatenated problems, of parameters mu_1 and mu_2, which can be treated independantly. If M = PDP', we consider Y_tilde = Y*P.

    dfmu_1 = 0:(2*n);
    dfmu_2 = 0:(n*(p-2));

    mu1 = dfmu_1;
    crit1 = mu1;
    mu2 = dfmu_2;
    crit2 = mu2;

    M =  J*J'/k+Jc*Jc'/(p-k);   
    [uM DM] = eig(M);
    eM = diag(DM);
    [~,bM] = sort(eM,'descend');
    uM = uM(:,bM);
    Y_prime= compY*(uM);
    Sigma_prime = uM'*S*uM;
    

% Problem n°1 : dimension 2, parameter mu_1

    Sigma_1 = Sigma_prime(1:2,1:2);
    kro = kron(Sigma_1,eye(n));
    
    for l = 1:length(mu1)
        if dfmu_1(l)>0
                     mu1(l) = df_to_lambda_eig(dfmu_1(l),eK_1);
        else
                    mu1(l) = Inf;
        end
        mu_1 = mu1(l);
        ecart = zeros(n*2,1);
        pen = 0;
        A_k = uK*diag( eK./(eK + n*2*mu_1*ones(n,1) ) )*uK';
        for k = 1:2
           ecart((k-1)*n+1:k*n,1) = (A_k-eye(n))*Y_prime(:,k);
           pen = pen + 2*trace(A_k*kro((k-1)*n+1:k*n,(k-1)*n+1:k*n));
        end    
        crit1(l) = (trace(ecart'*ecart) + pen);
    end    
    pos_1 = find_min(crit1);
    mu_1 = mu1(pos_1);
    
% Problem n°2 : dimension p-2, parameter mu_2

    Sigma_2 = Sigma_prime(3:p,3:p);
    kro = kron(Sigma_2,eye(n));

    for l = 1:length(mu2)
        if dfmu_2(l)>0
                     mu2(l) = df_to_lambda_eig(dfmu_2(l),eK_2);
        else
                    mu2(l) = Inf;
        end
        mu_2 = mu2(l);
        ecart = zeros(n*(p-2),1);
        pen = 0;
        A_k = uK*diag( eK./(eK + n*(p-2)*mu_2*ones(n,1) ) )*uK';
        for k = 3:p
            ecart((k-3)*n+1:(k-2)*n,1) = (A_k-eye(n))*Y_prime(:,k);
            pen = pen + 2*trace(A_k*kro((k-3)*n+1:(k-2)*n,(k-3)*n+1:(k-2)*n));
        end    
        crit2(l) = (trace(ecart'*ecart) + pen);
    end
   
    pos_2 = find_min(crit2);
    mu_2 = mu2(pos_2);
    crit_J(j) = crit1(pos_1)+crit2(pos_2);
    mus_J(:,j) = [mu_1;mu_2];
end

%%%% j = 2^p-1 or J = {1,...,p} 
dfmu_1 = 0:n;
dfmu_2 = 0:(n*(p-1));

mu1 = dfmu_1;
crit1 = mu1;
mu2 = dfmu_2; 
crit2 = mu2;

[uM DM] = eig(ones(p,p)/p);
eM = diag(DM);
[~,bM] = sort(eM,'descend');
uM = uM(:,bM);
Y_prime= compY*(uM);
Sigma_prime = uM'*S*uM;

% Problem n°1 : dim 1, parameter mu_1
y1 = Y_prime(:,1);
sigma1 = Sigma_prime(1,1);
for i = 1:length(dfmu_1)
    if dfmu_1(i)>0
       mu1(i) = df_to_lambda_eig(dfmu_1(i),eK);
    else
       mu1(i) = Inf;
    end
    mu_1 = mu1(i); 
    A = uK* diag( eK./(eK + n*mu_1*ones(n,1)) ) *uK';
    ecart = (eye(n)-A)*y1;
    crit1(i) = trace(ecart'*ecart) + 2*sigma1*trace(A);
end

pos_1 = find_min(crit1);
mus_J(1,2^p-1) = mu1(pos_1);

% Problem n°2 : dimension p-1, parameter mu_2

Sigma_2 = Sigma_prime(2:p,2:p);
kro = kron(Sigma_2,eye(n));

[~,eK_2] = eig_kron(eye(p-1),ones(p-1,1),uK',eK);
[~,bK_2] = sort(eK_2,'descend');
eK_2 = eK_2(bK_2);

for j = 1:length(mu2)
    if dfmu_2(j)>0
                 mu2(j) = df_to_lambda_eig(dfmu_2(j),eK_2);
         else
                 mu2(j) = Inf;
    end
    mu_2 = mu2(j);
    ecart = zeros(n*(p-1),1);
    pen = 0;
    A_k = uK*diag( eK./(eK + n*(p-1)*mu_2*ones(n,1) ) )*uK';
    for k = 2:p
        ecart((k-2)*n+1:(k-1)*n,1) = (A_k-eye(n))*Y_prime(:,k);
        pen = pen + 2*trace(A_k*kro((k-2)*n+1:(k-1)*n,(k-2)*n+1:(k-1)*n));
    end    
    crit2(j) = (trace(ecart'*ecart) + pen);
end    
pos_2 = find_min(crit2);

crit_J(2^p-1) = crit1(pos_1)+crit2(pos_2);
mus_J(2,2^p-1) = mu2(pos_2);

pos_j = find_min(crit_J);
mu_1 = mus_J(1,pos_j);
mu_2 = mus_J(2,pos_j);

J = str2num(sscanf(dec2bin(pos_j,p),'%c',[p,1]))';

vec_mu_1 = 1*((1:p) < 3)';
vec_mu_2 = 1*((1:p) > 2)';
mv1 =  multiply(mu_1,vec_mu_1);
mv2 =  multiply(mu_2,vec_mu_2);

e_invM = (mv1 + mv2).^(-1);
[uK_tilde,eK_tilde] = eig_kron(uM',e_invM,uK',eK);
A_chois = uK_tilde*diag(  eK_tilde./(eK_tilde+n*p*ones(n*p,1)) )*uK_tilde';
y_mult_pred = A_chois*y;

multi = (y_mult_pred - fX)'*(y_mult_pred - fX)/(n*p);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Single-task problem

Y_ind_pred = zeros(n,p);
for k = 1:p
    s = S(k,k);
    y_k = compY(:,k);
    critere = zeros(1,length(dfmu_1));
    for i = 1:length(dfmu_1)
        l = lambdas(i);
        A = uK* diag( eK./(eK + n*l*ones(n,1)) )*uK';
        pen = 2*s*trace(A);
        ecart = (eye(n) - A)*y_k;
        critere(i) = trace(ecart'*ecart)+pen;
     end
     i = find_min(critere); 
     l = lambdas(i);
     A_k = uK* diag( eK./(eK + n*l*ones(n,1)) )*uK';
     y_k_pred = A_k*y_k;
     Y_ind_pred(:,k) = y_k_pred;
end

y_ind_pred = zeros(n*p,1);
for k = 1:p
        y_ind_pred((k-1)*n+1:k*n) = Y_ind_pred(:,k);
end    

ind = (y_ind_pred-fX)'*(y_ind_pred-fX)/(n*p);

end    





